/*
 * js99'er - TI-99/4A emulator written in JavaScript
 *
 * Created 2014 by Rasmus Moustgaard <rasmus.moustgaard@gmail.com>
*/

 'use strict';

TI994A.FRAMES_TO_RUN = Number.MAX_VALUE;
TI994A.FRAME_MS = 16.667;
TI994A.FPS_MS = 2000;

function TI994A(document, canvas, diskImages) {

    // Assemble the console
    this.keyboard = new Keyboard();
    this.cru = new CRU(this.keyboard);
    this.tms9918 = new TMS9918(canvas.getContext("2d"), this.cru);
    var vdpRAM = this.tms9918.getRAM();
    this.diskDrives = [
        new DiskDrive("DSK1", vdpRAM, diskImages.FLOPPY1),
        new DiskDrive("DSK2", vdpRAM, diskImages.FLOPPY2),
        new DiskDrive("DSK3", vdpRAM, diskImages.FLOPPY3)
    ];
    this.tms9919 = new TMS9919();
    this.tms5220 = new TMS5220();
    this.memory = new Memory(this.tms9918, this.tms9919, this.tms5220);
    this.tms9900 = new TMS9900(this.memory, this.cru, this.diskDrives);
    this.cru.setMemory(this.memory);

    // Attach keyboard listeners
    var self = this;
    document.onkeydown = function(evt) {
        self.keyboard.keyEvent(evt, true);
    };
    document.onkeyup = function(evt) {
        self.keyboard.keyEvent(evt, false);
    };

    this.frameCount = 0;
    this.lastFpsTime = null;
    this.fpsFrameCount = 0;
    this.running = false;
    this.log = Log.getLog();

    this.reset();
}

TI994A.prototype = {

    isRunning: function() {
        return this.running;
    },

    reset: function() {
        this.tms9918.reset();
        this.tms9919.reset();
        this.keyboard.reset();
        this.memory.reset();
        this.cru.reset();
        this.tms9900.reset();
        this.resetFps();
    },

    start: function() {
        this.log.info("Start");
        var self = this;
        this.frameInterval = setInterval(
            function() {
                // Log.getLog().info("Frame " + self.frameCount);
                if (self.frameCount < TI994A.FRAMES_TO_RUN) {
                    self.frame();
                }
                else {
                    self.stop();
                }
            },
            TI994A.FRAME_MS
        );

        this.resetFps();
        this.printFps();
        this.fpsInterval = setInterval(
            function() {
                self.printFps();
            },
            TI994A.FPS_MS
        );
        this.running = true;
    },

    frame: function() {
        this.tms9900.run();
        this.tms9918.drawFrame();
        this.frameCount++;
        this.fpsFrameCount++;
    },

    stop: function() {
        this.log.info("Stop");
        clearInterval(this.frameInterval);
        clearInterval(this.fpsInterval);
        this.tms9919.mute();
        this.running = false;
    },

    resetFps: function() {
        this.lastFpsTime = null;
        this.fpsFrameCount = 0;
    },

    printFps: function() {
        var now = +new Date();
        var s = 'Frame ' + this.frameCount + ' running';
        if (this.lastFpsTime) {
            s += ': '
                + (this.fpsFrameCount / ((now - this.lastFpsTime) / 1000)).toFixed(1)
                + ' / '
                + (1000 / TI994A.FRAME_MS).toFixed(1)
                + ' FPS';
        }
        this.log.info(s);
        this.fpsFrameCount = 0;
        this.lastFpsTime = now;
    },

    getStatusString: function() {
        return "CPU: " + this.tms9900.getInternalRegsString() + "\n" +
               this.tms9900.getRegsStringFormatted() +
               this.memory.getStatusString() + "\n" +
               "VDP: " + this.tms9918.getRegsString();
    },

    getDiskDrives: function() {
        return this.diskDrives;
    },

    loadSoftware: function(sw) {
        var wasRunning = this.isRunning();
        if (wasRunning) {
            this.stop();
        }
        this.reset();
        if (sw.memoryBlocks != null) {
            for (var i = 0; i < sw.memoryBlocks.length; i++) {
                var memoryBlock = sw.memoryBlocks[i];
                this.memory.loadRAM(memoryBlock.address, memoryBlock.data);
            }
        }
        if (sw.rom != null) {
            this.memory.setCartridgeImage(sw.rom, sw.type == Software.TYPE_INVERTED_CART);
        }
        if (sw.grom != null) {
            this.memory.loadGROM(sw.grom, 3);
        }
        this.tms9900.setWP(sw.workspaceAddress != null ? sw.workspaceAddress : (SYSTEM.ROM[0] << 8 | SYSTEM.ROM[1]));
        this.tms9900.setPC(sw.startAddress != null ? sw.startAddress : (SYSTEM.ROM[2] << 8 | SYSTEM.ROM[3]));
        if (wasRunning) {
            this.start();
        }
    }
};
